/**
 * Copyright (C) 2011 Markus Gebhard (mk@jave.de) and others
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package de.jave.kicad.pcb.manipulate;

import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.List;

import net.disy.commons.core.util.CollectionUtilities;
import net.disy.commons.core.util.ITransformer;
import de.jave.kicad.pcb.model.Block;
import de.jave.kicad.pcb.model.DrawSegmentDescription;
import de.jave.kicad.pcb.model.DrawSegmentPosition;
import de.jave.kicad.pcb.model.FillSegment;
import de.jave.kicad.pcb.model.ModulePosition;
import de.jave.kicad.pcb.model.Pad;
import de.jave.kicad.pcb.model.PadDrill;
import de.jave.kicad.pcb.model.PadShape;
import de.jave.kicad.pcb.model.Placement;
import de.jave.kicad.pcb.model.PolysCorner;
import de.jave.kicad.pcb.model.TextPcbPosition;
import de.jave.kicad.pcb.model.TrackOrVia;
import de.jave.kicad.pcb.model.TrackOrViaDescription;
import de.jave.kicad.pcb.model.TrackOrViaPosition;
import de.jave.kicad.pcb.model.ZCorner;
import de.jave.kicad.pcb.persistence.IKicadConstants;
import de.jave.kicad.pcb.persistence.read.IBlockItem;
import de.jave.kicad.pcb.persistence.read.LineBlockItem;

//TODO_NOW (gebhard) 14.01.2011: make paramter codec?
public class BlockDecoder {

  public static String encodeRectangle(final Rectangle rectangle) {
    if (rectangle == null) {
      return "0 0 117000 82670"; //TODO_NOW (gebhard) 21.01.2011: take from sheet description: Sheet size!
    }
    return rectangle.x
        + " "
        + rectangle.y
        + " "
        + (rectangle.x + rectangle.width)
        + " "
        + (rectangle.y + rectangle.height);
  }

  public static Rectangle getRectangleParameter(final Block block, final String parameterName) {
    final String content = ParameterFinder.getRequiredParameterLineContent(block, parameterName);
    return parseRectangle(content);
  }

  private static Rectangle parseRectangle(final String string) {
    final String[] parts = splitIntoRequiredParts(string, 4);
    final int x1 = parseInt(parts[0]);
    final int y1 = parseInt(parts[1]);
    final int x2 = parseInt(parts[2]);
    final int y2 = parseInt(parts[3]);
    return new Rectangle(x1, y1, x2 - x1, y2 - y1);
  }

  private static String[] splitIntoRequiredParts(final String string, final int partCount) {
    final String[] parts = string.split(" "); //$NON-NLS-1$
    if (parts.length != partCount) {
      throw new IllegalArgumentException("Expected string to have " //$NON-NLS-1$
          + partCount
          + " parts, but were " //$NON-NLS-1$
          + parts.length
          + ": '"
          + string
          + "'"); //$NON-NLS-1$
    }
    return parts;
  }

  private static int parseInt(final String string) {
    return Integer.parseInt(string);
  }

  public static List<PolysCorner> getPolysCorners(final Block block) {
    final List<String> contents = ParameterFinder.getAllLineContents(block);
    return CollectionUtilities.transform(contents, new ITransformer<String, PolysCorner>() {
      @Override
      public PolysCorner transform(final String input) {
        return parsePolysCorner(input);
      }
    });
  }

  private static PolysCorner parsePolysCorner(final String string) {
    final String[] parts = splitIntoRequiredParts(string, 4);
    final int x = parseInt(parts[0]);
    final int y = parseInt(parts[1]);
    final int a = parseInt(parts[2]);
    final int b = parseInt(parts[3]);
    return new PolysCorner(x, y, a, b);
  }

  public static void setPolysCorners(final Block block, final List<PolysCorner> corners) {
    final List<String> contents = CollectionUtilities.transform(
        corners,
        new ITransformer<PolysCorner, String>() {
          @Override
          public String transform(final PolysCorner corner) {
            return encodePolysCorner(corner);
          }
        });
    ParameterFinder.replaceAllLines(block, contents);
  }

  private static String encodePolysCorner(final PolysCorner corner) {
    return corner.getX() + " " + corner.getY() + " " + corner.getA() + " " + corner.getB(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
  }

  public static List<ZCorner> getZCorners(final Block block) {
    final List<String> contents = ParameterFinder.getAllParameterLineContents(block, "ZCorner");
    return CollectionUtilities.transform(contents, new ITransformer<String, ZCorner>() {
      @Override
      public ZCorner transform(final String input) {
        return parseZCorner(input);
      }
    });
  }

  private static ZCorner parseZCorner(final String string) {
    final String[] parts = splitIntoRequiredParts(string, 3);
    final int x = parseInt(parts[0]);
    final int y = parseInt(parts[1]);
    final int flag = parseInt(parts[2]);//0: any corner, 1: Last corner
    return new ZCorner(x, y, flag);
  }

  public static void setZCorners(final Block block, final List<ZCorner> corners) {
    final List<String> contents = CollectionUtilities.transform(
        corners,
        new ITransformer<ZCorner, String>() {
          @Override
          public String transform(final ZCorner corner) {
            return encodeZCorner(corner);
          }
        });
    ParameterFinder.replaceParameterContent(block, "ZCorner", contents);
  }

  private static String encodeZCorner(final ZCorner corner) {
    return corner.getX() + " " + corner.getY() + " " + corner.getFlag(); //$NON-NLS-1$//$NON-NLS-2$
  }

  public static DrawSegmentPosition getDrawSegmentPosition(final Block block) {
    final String positionParameterContent = ParameterFinder.getRequiredParameterLineContent(
        block,
        "Po");
    return parseDrawSegmentPosition(positionParameterContent);
  }

  private static DrawSegmentPosition parseDrawSegmentPosition(final String string) {
    final String[] parts = splitIntoRequiredParts(string, 6);
    final int shape = parseInt(parts[0]);
    final int xStart = parseInt(parts[1]);
    final int yStart = parseInt(parts[2]);
    final int xEnd = parseInt(parts[3]);
    final int yEnd = parseInt(parts[4]);
    final int width = parseInt(parts[5]);
    return new DrawSegmentPosition(shape, xStart, yStart, xEnd, yEnd, width);
  }

  public static Block encodePad(final Pad pad) {
    final List<IBlockItem> items = new ArrayList<IBlockItem>();
    items
        .add(new LineBlockItem(IKicadConstants.PARAMETER_SH + ' ' + encodePadShape(pad.getShape())));
    items
        .add(new LineBlockItem(IKicadConstants.PARAMETER_DR + ' ' + encodePadDrill(pad.getDrill())));
    items.add(new LineBlockItem(IKicadConstants.PARAMETER_AT + ' ' + pad.getAttributesString()));
    items.add(new LineBlockItem(IKicadConstants.PARAMETER_NE + ' ' + pad.getNetString()));
    items.add(new LineBlockItem(IKicadConstants.PARAMETER_PO + ' ' + pad.getPositionString()));
    return new Block(IKicadConstants.TAG_PAD, items);
  }

  //TODO_NOW (gebhard) 26.01.2011: Extract to PadParser
  public static Pad parsePad(final Block block) {
    final PadShape padShape = parsePadShape(block);
    final PadDrill drill = parsePadDrill(block);

    //      At STD N 00E0FFFF
    final String attributesString = ParameterFinder.getRequiredParameterLineContent(
        block,
        IKicadConstants.PARAMETER_AT);

    //      Ne 3 "GND"
    final String netString = ParameterFinder.getRequiredParameterLineContent(
        block,
        IKicadConstants.PARAMETER_NE);

    //      Po -500 0
    final String positionString = ParameterFinder.getRequiredParameterLineContent(
        block,
        IKicadConstants.PARAMETER_PO);

    return new Pad(padShape, drill, attributesString, netString, positionString);
  }

  private static String encodePadDrill(final PadDrill drill) {
    if (drill.getShape() == null) {
      return drill.getDiameter() + " " + drill.getxOffset() + " " + drill.getyOffset(); //$NON-NLS-1$ //$NON-NLS-2$
    }
    return drill.getxDiameter() + " " //$NON-NLS-1$
        + drill.getxOffset()
        + " " //$NON-NLS-1$
        + drill.getyOffset()
        + " " //$NON-NLS-1$
        + drill.getShape()
        + " " //$NON-NLS-1$
        + drill.getxDiameter()
        + " " //$NON-NLS-1$
        + drill.getyDiameter();
  }

  //      Dr 1320 0 0 O 1320 320
  //      Dr 600 0 0
  private static PadDrill parsePadDrill(final Block block) {
    final String drillString = ParameterFinder.getRequiredParameterLineContent(
        block,
        IKicadConstants.PARAMETER_DR);
    final String[] parts = drillString.split(" "); //$NON-NLS-1$
    final int xOffset = parseInt(parts[1]);
    final int yOffset = parseInt(parts[2]);
    if (parts.length == 3) {
      final int diameter = parseInt(parts[0]);
      return new PadDrill(diameter, xOffset, yOffset);
    }
    else if (parts.length == 6) {
      final String shape = parts[3];
      final int xDiameter = parseInt(parts[4]);
      final int yDiameter = parseInt(parts[5]);
      return new PadDrill(xDiameter, yDiameter, xOffset, yOffset, shape);
    }
    else {
      throw new IllegalStateException();
    }
  }

  private static String encodePadShape(final PadShape shape) {
    return '"'
        + shape.getName()
        + '"'
        + ' '
        + shape.getShape()
        + ' '
        + shape.getxSize()
        + ' '
        + shape.getySize()
        + ' '
        + shape.getxDelta()
        + ' '
        + shape.getyDelta()
        + ' '
        + shape.getOrientation();
  }

  //      Sh "1" R 1700 700 0 0 2700
  private static PadShape parsePadShape(final Block block) {
    final String shapeString = ParameterFinder.getRequiredParameterLineContent(
        block,
        IKicadConstants.PARAMETER_SH);
    final int index0 = shapeString.indexOf('"');
    final int index1 = shapeString.indexOf('"', index0 + 1);
    final String padName = shapeString.substring(index0 + 1, index1);
    final String[] parts = splitIntoRequiredParts(shapeString.substring(index1 + 2), 6);
    final String shape = parts[0];
    final int xSize = parseInt(parts[1]);
    final int ySize = parseInt(parts[2]);
    final int xDelta = parseInt(parts[3]);
    final int yDelta = parseInt(parts[4]);
    final int orientation = parseInt(parts[5]);
    return new PadShape(padName, shape, xSize, ySize, xDelta, yDelta, orientation);
  }

  public static void setDrawSegmentPosition(final Block block, final DrawSegmentPosition position) {
    ParameterFinder.replaceParameterContent(block, "Po", encodeDrawSegmentPosition(position));
  }

  private static String encodeDrawSegmentPosition(final DrawSegmentPosition position) {
    return position.getShape() + " " //$NON-NLS-1$
        + position.getxStart()
        + " " //$NON-NLS-1$
        + position.getyStart()
        + " " //$NON-NLS-1$
        + position.getxEnd()
        + " "
        + position.getyEnd()
        + " " //$NON-NLS-1$
        + position.getWidth();
  }

  public static TextPcbPosition getTextPcbPosition(final Block block) {
    final String positionParameterContent = ParameterFinder.getRequiredParameterLineContent(
        block,
        "Po");
    return parseTextPcbPosition(positionParameterContent);
  }

  private static TextPcbPosition parseTextPcbPosition(final String string) {
    final String[] parts = splitIntoRequiredParts(string, 6);
    final int xStart = parseInt(parts[0]);
    final int yStart = parseInt(parts[1]);
    final int xSize = parseInt(parts[2]);
    final int ySize = parseInt(parts[3]);
    final int width = parseInt(parts[4]);
    final int rotation = parseInt(parts[5]);
    return new TextPcbPosition(xStart, yStart, xSize, ySize, width, rotation);
  }

  public static void setTextPcbPosition(final Block block, final TextPcbPosition position) {
    ParameterFinder.replaceParameterContent(block, "Po", encodeTextPcbPosition(position));
  }

  private static String encodeTextPcbPosition(final TextPcbPosition position) {
    return position.getxStart() + " " //$NON-NLS-1$
        + position.getyStart()
        + " " //$NON-NLS-1$
        + position.getxSize()
        + " " //$NON-NLS-1$
        + position.getySize()
        + " " //$NON-NLS-1$
        + position.getWidth()
        + " " //$NON-NLS-1$
        + position.getRotation();
  }

  public static ModulePosition getModulePosition(final Block block) {
    final String positionParameterContent = ParameterFinder.getRequiredParameterLineContent(
        block,
        "Po");
    return parseModulePosition(positionParameterContent);
  }

  public static void setModulePosition(final Block block, final ModulePosition position) {
    ParameterFinder.replaceParameterContent(block, "Po", encodeModulePosition(position));
  }

  private static String encodeModulePosition(final ModulePosition position) {
    return position.getX() + " " //$NON-NLS-1$
        + position.getY()
        + " " //$NON-NLS-1$
        + position.getOrientation()
        + " " //$NON-NLS-1$
        + position.getLayer()
        + " " //$NON-NLS-1$
        + position.getS1()
        + " " //$NON-NLS-1$
        + position.getS2()
        + " " //$NON-NLS-1$
        + enocePlacement(position.getPlacement());
  }

  private static String enocePlacement(final Placement placement) {
    switch (placement) {
      case AUTOPLACED:
        return "~P";
      case FIXED:
        return "F~";
      case FIXED_AUTOPLACED:
        return "FP";
      case MOVEABLE:
        return "~~";
      default:
        throw new UnsupportedOperationException();
    }
  }

  private static ModulePosition parseModulePosition(final String string) {
    final String[] parts = splitIntoRequiredParts(string, 7);
    final int x = parseInt(parts[0]);
    final int y = parseInt(parts[1]);
    final int orientation = parseInt(parts[2]);
    final int layer = parseInt(parts[3]);
    final String s1 = parts[4];
    final String s2 = parts[5];
    final Placement placement = parsePlacement(parts[6]);
    return new ModulePosition(x, y, orientation, layer, s1, s2, placement);
  }

  private static Placement parsePlacement(final String string) {
    if ("~~".equals(string)) {
      return Placement.MOVEABLE;
    }
    if ("F~".equals(string)) {
      return Placement.FIXED;
    }
    if ("~P".equals(string)) {
      return Placement.AUTOPLACED;
    }
    if ("FP".equals(string)) {
      return Placement.FIXED_AUTOPLACED;
    }
    throw new IllegalArgumentException(string);
  }

  public static List<FillSegment> getFillSegments(final Block block) {
    final List<String> contents = ParameterFinder.getAllLineContents(block);
    return CollectionUtilities.transform(contents, new ITransformer<String, FillSegment>() {
      @Override
      public FillSegment transform(final String input) {
        return parseFillSegment(input);
      }
    });
  }

  private static FillSegment parseFillSegment(final String string) {
    final String[] parts = splitIntoRequiredParts(string, 4);
    final int xStart = parseInt(parts[0]);
    final int yStart = parseInt(parts[1]);
    final int xEnd = parseInt(parts[2]);
    final int yEnd = parseInt(parts[3]);
    return new FillSegment(xStart, yStart, xEnd, yEnd);
  }

  public static void setFillSegments(final Block block, final List<FillSegment> segments) {
    final List<String> contents = CollectionUtilities.transform(
        segments,
        new ITransformer<FillSegment, String>() {
          @Override
          public String transform(final FillSegment segment) {
            return encodeFillSegment(segment);
          }
        });
    ParameterFinder.replaceAllLines(block, contents);
  }

  private static String encodeFillSegment(final FillSegment segment) {
    return segment.getxStart() + " " //$NON-NLS-1$
        + segment.getyStart()
        + " " //$NON-NLS-1$
        + segment.getxEnd()
        + " "
        + segment.getyEnd();
  }

  public static List<TrackOrVia> getTracksOrVias(final Block trackBlock) {
    final List<TrackOrVia> list = new ArrayList<TrackOrVia>();
    final List<IBlockItem> items = trackBlock.getItems();
    int index = 0;
    while (index < items.size()) {
      String position = null;
      while (index < items.size()) {
        final IBlockItem item = items.get(index++);
        if (item instanceof LineBlockItem) {
          final LineBlockItem lineItem = (LineBlockItem) item;
          if (lineItem.getLine().startsWith("Po ")) {
            position = lineItem.getLine().substring("Po ".length());
            break;
          }
        }
      }
      if (position == null) {
        break;
      }
      String description = null;
      while (index < items.size()) {
        final IBlockItem item = items.get(index++);
        if (item instanceof LineBlockItem) {
          final LineBlockItem lineItem = (LineBlockItem) item;
          if (lineItem.getLine().startsWith("De ")) {
            description = lineItem.getLine().substring("De ".length());
            break;
          }
        }
      }
      if (description == null) {
        break;
      }
      final TrackOrVia tov = parseTrackOrVia(position, description);
      list.add(tov);
    }

    return list;
  }

  private static TrackOrVia parseTrackOrVia(final String position, final String description) {
    final TrackOrViaPosition trackOrViaPosition = parseTrackOrViaPosition(position);
    final TrackOrViaDescription trackOrViaDescription = parseTrackOrViaDescription(description);
    return new TrackOrVia(trackOrViaPosition, trackOrViaDescription);
  }

  private static TrackOrViaDescription parseTrackOrViaDescription(final String description) {
    final String[] parts = splitIntoRequiredParts(description, 5);
    final int layer = parseInt(parts[0]);
    final int type = parseInt(parts[1]);
    final int netcode = parseInt(parts[2]);
    final int timestamp = parseInt(parts[3]);
    final String status = parts[4];
    return new TrackOrViaDescription(layer, type, netcode, timestamp, status);
  }

  public static String encodeTrackOrViaDescription(final TrackOrViaDescription description) {
    return "De " + description.getLayer() + " " //$NON-NLS-1$ //$NON-NLS-2$
        + description.getType()
        + " " //$NON-NLS-1$
        + description.getNetcode()
        + " " //$NON-NLS-1$
        + description.getTimestamp()
        + " " //$NON-NLS-1$
        + description.getStatus();
  }

  private static TrackOrViaPosition parseTrackOrViaPosition(final String position) {
    final String[] parts = splitIntoRequiredParts(position, 7);
    final int shape = parseInt(parts[0]);
    final int xStart = parseInt(parts[1]);
    final int yStart = parseInt(parts[2]);
    final int xEnd = parseInt(parts[3]);
    final int yEnd = parseInt(parts[4]);
    final int widthOrDiameter = parseInt(parts[5]);
    final int a = parseInt(parts[6]);
    return new TrackOrViaPosition(shape, xStart, yStart, xEnd, yEnd, widthOrDiameter, a);
  }

  public static String encodeTrackOrViaPosition(final TrackOrViaPosition position) {
    return "Po " + position.getShape() + " " //$NON-NLS-1$ //$NON-NLS-2$
        + position.getxStart()
        + " " //$NON-NLS-1$
        + position.getyStart()
        + " " //$NON-NLS-1$
        + position.getxEnd()
        + " " //$NON-NLS-1$
        + position.getyEnd()
        + " " //$NON-NLS-1$
        + position.getWidthOrDiameter()
        + " " //$NON-NLS-1$
        + position.getA();
  }

  public static DrawSegmentDescription getDrawSegmentDescription(final Block block) {
    final String content = ParameterFinder.getRequiredParameterLineContent(block, "De");
    final String[] parts = splitIntoRequiredParts(content, 5);
    final int layer = parseInt(parts[0]);
    final int type = parseInt(parts[1]);
    final int angle = parseInt(parts[2]);
    final int timestamp = parseInt(parts[3]);
    final String status = parts[4];
    return new DrawSegmentDescription(layer, type, angle, timestamp, status);
  }
}