/**
 * Copyright (C) 2011 Markus Gebhard (mk@jave.de) and others
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package de.jave.kicad.pcb.model;

import java.awt.Rectangle;
import java.io.File;
import java.util.ArrayList;
import java.util.List;

import net.disy.commons.core.util.CollectionUtilities;
import net.disy.commons.core.util.Ensure;
import net.disy.commons.core.util.ITransformer;
import de.jave.kicad.pcb.manipulate.BlockDecoder;
import de.jave.kicad.pcb.manipulate.BlockFinder;
import de.jave.kicad.pcb.manipulate.ParameterFinder;
import de.jave.kicad.pcb.persistence.IKicadConstants;
import de.jave.kicad.pcb.persistence.read.BlockBlockItem;
import de.jave.kicad.pcb.persistence.read.IBlockItem;
import de.jave.kicad.pcb.persistence.read.LineBlockItem;

public class KicadBoard implements Cloneable {

  private final Block rootBlock;
  private File file;

  public KicadBoard(final File file, final Block rootBlock) {
    Ensure.ensureArgumentNotNull(file);
    Ensure.ensureArgumentNotNull(rootBlock);
    this.file = file;
    this.rootBlock = rootBlock;
  }

  @Override
  public KicadBoard clone() {
    return new KicadBoard(file, rootBlock.clone());
  }

  @Deprecated
  public Block getRootBlock() {
    return rootBlock;
  }

  private void updateBounds() {
    final Block generalBlock = BlockFinder.getFirstRequiredBlock(rootBlock, "GENERAL");
    final Rectangle newBounds = getContent().getMaxBounds();
    ParameterFinder.replaceParameterContent(
        generalBlock,
        "Di",
        BlockDecoder.encodeRectangle(newBounds));
  }

  private TracksAndVias getTracksAndVias() {
    final Block trackBlock = BlockFinder.getFirstRequiredBlock(rootBlock, "TRACK"); //$NON-NLS-1$
    return new TracksAndVias(BlockDecoder.getTracksOrVias(trackBlock));
  }

  private void setTracksAndVias(final TracksAndVias tracksOrVias) {
    removeTracksAndVis();
    addTracksAndVias(tracksOrVias);
  }

  private void removeTracksAndVis() {
    removeAllBlocksByPrefix(IKicadConstants.TAG_TRACK);
    rootBlock.getItems().add(
        new BlockBlockItem(new Block(IKicadConstants.TAG_TRACK, new ArrayList<IBlockItem>())));
  }

  private void addTracksAndVias(final TracksAndVias newTracksOrVias) {
    final List<String> lines = enocde(newTracksOrVias.getEntries());
    Block block = BlockFinder.getFirstBlock(rootBlock, "TRACK"); //$NON-NLS-1$
    if (block == null) {
      block = new Block("TRACK", new ArrayList<IBlockItem>());
      rootBlock.getItems().add(new BlockBlockItem(block));
    }
    final List<IBlockItem> items = block.getItems();
    for (final String line : lines) {
      items.add(new LineBlockItem(line));
    }
    updateBounds();
  }

  private List<String> enocde(final List<TrackOrVia> newTracksOrVias) {
    final List<String> lines = new ArrayList<String>();
    for (final TrackOrVia trackOrVia : newTracksOrVias) {
      lines.add(BlockDecoder.encodeTrackOrViaPosition(trackOrVia.getPosition()));
      lines.add(BlockDecoder.encodeTrackOrViaDescription(trackOrVia.getDescription()));
    }
    return lines;
  }

  private ModuleList getModules() {
    final List<Block> moduleBlocks = BlockFinder.getAllBlocksByPrefix(
        rootBlock,
        IKicadConstants.TAG_MODULE);
    final List<Module> modules = CollectionUtilities.transform(
        moduleBlocks,
        new ITransformer<Block, Module>() {
          @Override
          public Module transform(final Block input) {
            return new Module(input);
          }
        });
    return new ModuleList(modules);
  }

  private void setModules(final ModuleList modules) {
    removeModules();
    addModules(modules);
  }

  private void removeModules() {
    removeAllBlocksByPrefix(IKicadConstants.TAG_MODULE);
  }

  private void addModules(final ModuleList modules) {
    rootBlock.getItems().addAll(modules.getBlockItems());
    updateBounds();
  }

  private DrawSegmentList getDrawSegments() {
    final List<Block> drawSegmentBlocks = BlockFinder.getAllBlocksByPrefix(
        rootBlock,
        IKicadConstants.TAG_DRAWSEGMENT);
    final List<DrawSegment> drawSegments = CollectionUtilities.transform(
        drawSegmentBlocks,
        new ITransformer<Block, DrawSegment>() {
          @Override
          public DrawSegment transform(final Block input) {
            return new DrawSegment(input);
          }
        });
    return new DrawSegmentList(drawSegments);
  }

  private void setDrawSegments(final DrawSegmentList drawSegments) {
    removeDrawSegments();
    addDrawSegments(drawSegments);
  }

  private void removeDrawSegments() {
    removeAllBlocksByPrefix(IKicadConstants.TAG_DRAWSEGMENT);
  }

  private void addDrawSegments(final DrawSegmentList drawSegments) {
    rootBlock.getItems().addAll(drawSegments.getBlockItems());
    updateBounds();
  }

  private TextList getTexts() {
    final List<Block> textPcbBlocks = BlockFinder.getAllBlocksByPrefix(
        rootBlock,
        IKicadConstants.TAG_TEXTPCB);
    final List<Text> texts = CollectionUtilities.transform(
        textPcbBlocks,
        new ITransformer<Block, Text>() {
          @Override
          public Text transform(final Block input) {
            return new Text(input);
          }
        });
    return new TextList(texts);
  }

  private void setTexts(final TextList texts) {
    removeTexts();
    addTexts(texts);
  }

  private void removeTexts() {
    removeAllBlocksByPrefix(IKicadConstants.TAG_TEXTPCB);
  }

  private void addTexts(final TextList texts) {
    rootBlock.getItems().addAll(texts.getBlockItems());
    updateBounds();
  }

  private ZoneList getZones() {
    final List<Block> outlineBlocks = BlockFinder.getAllBlocksByPrefix(
        rootBlock,
        IKicadConstants.TAG_CZONE_OUTLINE);
    final List<Zone> zones = CollectionUtilities.transform(
        outlineBlocks,
        new ITransformer<Block, Zone>() {
          @Override
          public Zone transform(final Block input) {
            return new Zone(input);
          }
        });
    return new ZoneList(zones);
  }

  private void setZones(final ZoneList zones) {
    removeZones();
    addZones(zones);
  }

  private void removeZones() {
    removeAllBlocksByPrefix(IKicadConstants.TAG_CZONE_OUTLINE);
  }

  private void addZones(final ZoneList zones) {
    rootBlock.getItems().addAll(zones.getBlockItems());
    updateBounds();
  }

  public KicadBoardContent getContent() {
    final ModuleList modules = getModules();
    final DrawSegmentList drawSegments = getDrawSegments();
    final TextList texts = getTexts();
    final ZoneList zones = getZones();
    final TracksAndVias tracksOrVias = getTracksAndVias();
    return new KicadBoardContent(modules, drawSegments, texts, zones, tracksOrVias);
  }

  public void setContent(final KicadBoardContent content) {
    setModules(content.getModules());
    setDrawSegments(content.getDrawSegments());
    setTexts(content.getTexts());
    setTracksAndVias(content.getTracksOrVias());
    setZones(content.getZones());
  }

  public void removeContent() {
    removeModules();
    removeDrawSegments();
    removeTexts();
    removeTracksAndVis();
    removeZones();
    updateBounds();
  }

  public void addContent(final KicadBoardContent content) {
    addModules(content.getModules());
    addDrawSegments(content.getDrawSegments());
    addTexts(content.getTexts());
    addTracksAndVias(content.getTracksOrVias());
    addZones(content.getZones());
  }

  private void removeAllBlocksByPrefix(final String prefix) {
    final List<BlockBlockItem> oldItems = BlockFinder.getAllBlockItemsByPrefix(rootBlock, prefix);
    rootBlock.getItems().removeAll(oldItems);
  }

  public Rectangle getPcbOutlineBounds() {
    return getDrawSegments().getPcbEdgesBounds();
  }

  public File getFile() {
    return file;
  }

  public void setFile(final File file) {
    Ensure.ensureArgumentNotNull(file);
    this.file = file;
  }

  public Rectangle getPcbOutlineOrContentBounds() {
    final Rectangle outlineBounds = getPcbOutlineBounds();
    if (outlineBounds != null) {
      return outlineBounds;
    }
    return getContent().getMaxBounds();
  }
}