/**
 * Copyright (C) 2011 Markus Gebhard (mk@jave.de) and others
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package de.jave.kicad.pcbmultiplyer;

import java.awt.Dimension;
import java.util.ArrayList;
import java.util.List;

import net.disy.commons.core.progress.ICancelable;
import net.disy.commons.core.progress.ProgressUtilities;
import net.disy.commons.core.util.CollectionUtilities;
import net.disy.commons.core.util.Ensure;
import net.disy.commons.core.util.ITransformer;
import de.jave.kicad.pcbmultiplyer.tree.AddHorizontalSolution;
import de.jave.kicad.pcbmultiplyer.tree.AddVerticalSolution;
import de.jave.kicad.pcbmultiplyer.tree.ISolution;
import de.jave.kicad.pcbmultiplyer.tree.MultiplySolution;
import de.jave.kicad.pcbmultiplyer.tree.PcbSolution;
import de.jave.kicad.pcbmultiplyer.tree.RotateSolution;

public class SolutionFactory {

  private final PcbMultiplySetup setup;
  private final SolutionRater solutionRater;

  public SolutionFactory(final PcbMultiplySetup setup, final SolutionRater solutionRater) {
    Ensure.ensureArgumentNotNull(setup);
    Ensure.ensureArgumentNotNull(solutionRater);
    this.setup = setup;
    this.solutionRater = solutionRater;
  }

  public ISolution createBestSolutionWithDepth(final ICancelable cancelable)
      throws InterruptedException {
    final List<ISolution> solutions = createAllSolutions(cancelable);
    if (solutions.isEmpty()) {
      return null;
    }
    ISolution bestSolution = solutions.get(0);
    for (int i = 0; i < solutions.size(); ++i) {
      ProgressUtilities.checkInterrupted(cancelable);
      final ISolution currentSolution = solutions.get(i);
      if (bestSolution == null
          || solutionRater.rate(currentSolution) > solutionRater.rate(bestSolution)) {
        bestSolution = currentSolution;
      }
    }
    return bestSolution;
  }

  private List<ISolution> createAllSolutions(final ICancelable cancelable)
      throws InterruptedException {

    final int targetWidth = setup.getTargetBoardBounds().width;
    final int targetHeight = setup.getTargetBoardBounds().height;

    final Dimension contentSize = setup.getOriginalContentBounds().getSize();
    final int contentHeight = contentSize.height;
    final int contentWidth = contentSize.width;

    final int spacing = setup.getSpacing();

    final List<ISolution> solutions = createVerticallySplitSolutions(
        cancelable,
        targetWidth,
        targetHeight,
        contentSize,
        contentHeight,
        contentWidth,
        spacing);

    final List<ISolution> rotatedTargetBoardSolutions = createVerticallySplitSolutions(
        cancelable,
        targetHeight,
        targetWidth,
        contentSize,
        contentHeight,
        contentWidth,
        spacing);

    solutions.addAll(CollectionUtilities.transform(
        rotatedTargetBoardSolutions,
        new ITransformer<ISolution, ISolution>() {
          @Override
          public ISolution transform(final ISolution input) {
            return new RotateSolution(input);
          }
        }));
    return solutions;
  }

  private List<ISolution> createVerticallySplitSolutions(
      final ICancelable cancelable,
      final int targetWidth,
      final int targetHeight,
      final Dimension contentSize,
      final int contentHeight,
      final int contentWidth,
      final int spacing) throws InterruptedException {
    final List<ISolution> solutions = new ArrayList<ISolution>();
    int columnsLeft = 0;
    while (true) {
      ProgressUtilities.checkInterrupted(cancelable);
      final int rowsLeft = getPossibleMultiples(targetHeight, contentHeight, spacing);
      final PcbSolution pcbSolution = new PcbSolution(contentSize);
      ISolution leftSolution = new MultiplySolution(pcbSolution, columnsLeft, rowsLeft, spacing);
      if (leftSolution.getSize().width > targetWidth) {
        break;
      }
      final int rowsRight = getPossibleMultiples(targetHeight, contentWidth, spacing);
      final int remainingWidth = targetWidth - leftSolution.getSize().width - spacing;
      final int columnsRight = getPossibleMultiples(remainingWidth, contentHeight, spacing);
      ISolution rightSolution = new MultiplySolution(
          new RotateSolution(pcbSolution),
          columnsRight,
          rowsRight,
          spacing);

      if (rightSolution.getSize().height + spacing + contentHeight <= targetHeight) {
        final int remainingTopWidth = targetWidth - spacing - leftSolution.getSize().width;
        final int columnCount = getPossibleMultiples(remainingTopWidth, contentWidth, spacing);
        if (columnCount > 0) {
          final ISolution topSolution = new MultiplySolution(pcbSolution, columnCount, 1, spacing);
          rightSolution = new AddVerticalSolution(rightSolution, topSolution, spacing);
        }
      }
      else if (rightSolution.getSize().height + spacing + contentWidth <= targetHeight) {
        final int remainingTopWidth = targetWidth - spacing - leftSolution.getSize().width;
        final int columnCount = getPossibleMultiples(remainingTopWidth, contentHeight, spacing);
        if (columnCount > 0) {
          final ISolution topSolution = new MultiplySolution(
              new RotateSolution(pcbSolution),
              columnCount,
              1,
              spacing);
          rightSolution = new AddVerticalSolution(rightSolution, topSolution, spacing);
        }
      }
      else if (leftSolution.getSize().height + spacing + contentHeight <= targetHeight) {
        final int remainingTopWidth = targetWidth - spacing - rightSolution.getSize().width;
        final int columnCount = getPossibleMultiples(remainingTopWidth, contentWidth, spacing);
        if (columnCount > 0) {
          final ISolution topSolution = new MultiplySolution(pcbSolution, columnCount, 1, spacing);
          leftSolution = new AddVerticalSolution(rightSolution, topSolution, spacing);
        }
      }
      else if (leftSolution.getSize().height + spacing + contentWidth <= targetHeight) {
        final int remainingTopWidth = targetWidth - spacing - leftSolution.getSize().width;
        final int columnCount = getPossibleMultiples(remainingTopWidth, contentHeight, spacing);
        if (columnCount > 0) {
          final ISolution topSolution = new MultiplySolution(
              new RotateSolution(pcbSolution),
              columnCount,
              1,
              spacing);
          leftSolution = new AddVerticalSolution(rightSolution, topSolution, spacing);
        }
      }

      final AddHorizontalSolution solution = new AddHorizontalSolution(
          leftSolution,
          rightSolution,
          spacing);
      final AddHorizontalSolution mirroredSolution = new AddHorizontalSolution(
          rightSolution,
          leftSolution,
          spacing);
      solutions.add(solution);
      solutions.add(mirroredSolution);
      ++columnsLeft;
    }
    return solutions;
  }

  private int getPossibleMultiples(final int totalLength, final int length, final int spacing) {
    if (length > totalLength) {
      return 0;
    }
    return 1 + (totalLength - length) / (length + spacing);
  }
}