/**
 * Copyright (C) 2011 Markus Gebhard (mk@jave.de) and others
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package de.jave.kicad.pcbmultiplyer.preferences;

import java.io.File;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

import net.disy.commons.core.io.FileModel;
import net.disy.commons.core.model.listener.IChangeListener;
import de.jave.kicad.pcbmultiplyer.MultiplyMode;
import de.jave.kicad.pcbmultiplyer.PcbMultiplyOptionsModel;

public class PcbMultiplyPreferences {

  private static final String ROOT_KEY = "de.jave.kicad.pcbmultiplyer"; //$NON-NLS-1$
  private static final String KEY_BOARD_HEIGHT_MM = "boardHeightMm"; //$NON-NLS-1$
  private static final String KEY_BOARD_PRICE = "boardPrice"; //$NON-NLS-1$
  private static final String KEY_BOARD_WIDTH_MM = "boardWidthMm"; //$NON-NLS-1$
  private static final String KEY_CURRENT_DIRECTORY = "currentDirectory"; //$NON-NLS-1$
  private static final String KEY_SPACING_MM = "spacingMm"; //$NON-NLS-1$
  private static final String KEY_MULTIPLES_X = "multiplesX"; //$NON-NLS-1$
  private static final String KEY_MULTIPLES_Y = "multiplesY"; //$NON-NLS-1$
  private static final String KEY_MODE = "mode"; //$NON-NLS-1$

  private final Preferences preferences;
  private final FileModel currentDirectoryModel = new FileModel();
  private final PcbMultiplyOptionsModel optionsModel = new PcbMultiplyOptionsModel();

  public PcbMultiplyPreferences() {
    preferences = Preferences.userRoot().node(ROOT_KEY);
    initializeCurrentDirectoryModel();
    initializeOptionsModel();
  }

  private void initializeOptionsModel() {
    final int boardWidthMm = getInt(KEY_BOARD_WIDTH_MM, optionsModel.getBoardWidthMm());
    optionsModel.setBoardWidthMm(boardWidthMm);

    final int boardHeightMm = getInt(KEY_BOARD_HEIGHT_MM, optionsModel.getBoardHeightMm());
    optionsModel.setBoardHeightMm(boardHeightMm);

    final int boardPrice = getInt(KEY_BOARD_PRICE, optionsModel.getBoardPrice());
    optionsModel.setBoardPrice(boardPrice);

    final double spacingMm = getDouble(KEY_SPACING_MM, optionsModel.getSpacingMm());
    optionsModel.setSpacingMm(spacingMm);

    final int multiplesX = getInt(KEY_MULTIPLES_X, optionsModel.getMultiplesX());
    optionsModel.setMultiplesX(multiplesX);

    final int multiplesY = getInt(KEY_MULTIPLES_Y, optionsModel.getMultiplesY());
    optionsModel.setMultiplesY(multiplesY);

    final int modeId = getInt(KEY_MODE, optionsModel.getMode().getPersistenceId());
    final MultiplyMode mode = MultiplyMode.getByPersistenceId(modeId);
    if (mode != null) {
      optionsModel.setMode(mode);
    }
  }

  private void initializeCurrentDirectoryModel() {
    final String directory = get(KEY_CURRENT_DIRECTORY, null);
    if (directory != null) {
      currentDirectoryModel.setValue(new File(directory));
    }
    currentDirectoryModel.addChangeListener(new IChangeListener() {
      @Override
      public void stateChanged() {
        final File currentDirectory = currentDirectoryModel.getValue();
        put(KEY_CURRENT_DIRECTORY, currentDirectory);
        flush();
      }
    });
  }

  private void put(final String key, final File file) {
    put(key, file == null ? (String) null : file.getAbsolutePath());
  }

  private final void put(final String key, final String value) {
    if (value == null) {
      preferences.remove(key);
    }
    else {
      preferences.put(key, value);
    }
  }

  private final void putInt(final String key, final int value) {
    preferences.putInt(key, value);
  }

  private final void putDouble(final String key, final double value) {
    preferences.putDouble(key, value);
  }

  private final String get(final String key, final String defaultValue) {
    return preferences.get(key, defaultValue);
  }

  private final int getInt(final String key, final int defaultValue) {
    return preferences.getInt(key, defaultValue);
  }

  private final double getDouble(final String key, final double defaultValue) {
    return preferences.getDouble(key, defaultValue);
  }

  public FileModel getCurrentDirectoryModel() {
    return currentDirectoryModel;
  }

  public void storeOptions() {
    putInt(KEY_BOARD_WIDTH_MM, optionsModel.getBoardWidthMm());
    putInt(KEY_BOARD_HEIGHT_MM, optionsModel.getBoardHeightMm());
    putInt(KEY_BOARD_PRICE, optionsModel.getBoardPrice());
    putDouble(KEY_SPACING_MM, optionsModel.getSpacingMm());
    putInt(KEY_MULTIPLES_X, optionsModel.getMultiplesX());
    putInt(KEY_MULTIPLES_Y, optionsModel.getMultiplesY());
    putInt(KEY_MODE, optionsModel.getMode().getPersistenceId());
  }

  public final void flush() {
    try {
      preferences.flush();
    }
    catch (final BackingStoreException exception) {
      //We ignore if saving fales - just print the stacktrace.
      exception.printStackTrace();
    }
  }

  public PcbMultiplyOptionsModel getOptionsModel() {
    return optionsModel;
  }
}