/**
 * Copyright (C) 2011 Markus Gebhard (mk@jave.de) and others
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package de.jave.kicad.pcbmultiplyer.swing;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.text.DecimalFormat;
import java.util.List;

import javax.swing.JComponent;

import net.disy.commons.core.model.ObjectModel;
import net.disy.commons.core.model.listener.IChangeListener;
import net.disy.commons.core.util.Ensure;
import de.jave.kicad.pcbmultiplyer.OrientedPolygon;
import de.jave.kicad.pcbmultiplyer.PcbMultiplySetup;
import de.jave.kicad.pcbmultiplyer.SolutionPolygonFactory;
import de.jave.kicad.pcbmultiplyer.tree.ISolution;

public class SolutionComponent extends JComponent {

  private static final DecimalFormat MM_FORMAT = new DecimalFormat("#mm"); //$NON-NLS-1$
  private static final Color BACKGROUND_COLOR = Color.BLACK;
  private static final Color BOARD_COLOR = new Color(165, 156, 49);
  private static final Color BOARD_SIZE_TEXT_COLOR = BOARD_COLOR.brighter();
  private static final Color BOARD_OUTLINE_COLOR = BOARD_COLOR.darker();
  private static final Color PCB_OUTLINE_COLOR = Color.YELLOW;
  private static final Color PCB_FILL_COLOR = new Color(0, 160, 0, 128);

  private static final int BORDER_SIZE = 20;

  private final ObjectModel<PcbMultiplySetup> setupModel;
  private final ObjectModel<ISolution> solutionModel;

  public SolutionComponent(
      final ObjectModel<PcbMultiplySetup> setupModel,
      final ObjectModel<ISolution> solutionModel) {
    Ensure.ensureArgumentNotNull(setupModel);
    Ensure.ensureArgumentNotNull(solutionModel);
    this.setupModel = setupModel;
    this.solutionModel = solutionModel;
    final IChangeListener repaintListener = new IChangeListener() {
      @Override
      public void stateChanged() {
        repaint();
      }
    };
    solutionModel.addChangeListener(repaintListener);
    setupModel.addChangeListener(repaintListener);
  }

  @Override
  public Dimension getPreferredSize() {
    return new Dimension(640, 480);
  }

  @Override
  protected void paintComponent(final Graphics g) {
    renderBackground(g);

    final PcbMultiplySetup setup = setupModel.getValue();
    if (setup == null) {
      return;
    }
    g.setFont(new Font(Font.DIALOG, Font.PLAIN, 11));
    final IRenderTransformation t = getRenderTransformation(setup);
    renderBoard(g, t, setup);
    final ISolution solution = solutionModel.getValue();
    if (solution == null) {
      return;
    }
    renderPcbBounds(g, t, setup, solution);
  }

  private void renderBackground(final Graphics g) {
    g.setColor(BACKGROUND_COLOR);
    g.fillRect(0, 0, getWidth(), getHeight());
  }

  private void renderBoard(
      final Graphics g,
      final IRenderTransformation t,
      final PcbMultiplySetup setup) {
    final Rectangle boardBounds = setup.getTargetBoardBounds();
    g.setColor(BOARD_COLOR);
    final int x1 = t.transformX(boardBounds.x);
    final int y1 = t.transformY(boardBounds.y);
    final int x2 = t.transformX(boardBounds.x + boardBounds.width);
    final int y2 = t.transformY(boardBounds.y + boardBounds.height);
    g.fillRect(x1, y1, x2 - x1, y2 - y1);
    g.setColor(BOARD_OUTLINE_COLOR);
    g.drawRect(x1, y1, x2 - x1, y2 - y1);

    g.setColor(BOARD_SIZE_TEXT_COLOR);
    final String widthText = MM_FORMAT
        .format((double) setup.getTargetBoardBounds().width * 254 / 100000);
    final FontMetrics fontMetrics = g.getFontMetrics();
    final int widthTextWidth = fontMetrics.stringWidth(widthText);
    g.drawString(widthText, (x1 + x2) / 2 - widthTextWidth / 2, y1 - fontMetrics.getDescent());

    final String heightText = MM_FORMAT
        .format((double) setup.getTargetBoardBounds().height * 254 / 100000);
    final int heightTextWidth = fontMetrics.stringWidth(heightText);

    final Graphics2D tmpGraphics = (Graphics2D) g.create();
    tmpGraphics.translate(x1, (y1 + y2) / 2);
    tmpGraphics.rotate(-Math.PI / 2);
    tmpGraphics.drawString(heightText, -heightTextWidth / 2, 0 - fontMetrics.getDescent());
    tmpGraphics.dispose();
  }

  private void renderPcbBounds(
      final Graphics g,
      final IRenderTransformation t,
      final PcbMultiplySetup setup,
      final ISolution solution) {
    final List<OrientedPolygon> pcbBounds = SolutionPolygonFactory.createPolygons(setup, solution);
    for (int i = 0; i < pcbBounds.size(); ++i) {
      final OrientedPolygon polygon = pcbBounds.get(i);
      final Polygon renderPolygon = transform(polygon.getPolygon(), t);
      g.setColor(PCB_FILL_COLOR);
      g.fillPolygon(renderPolygon);
      g.setColor(PCB_OUTLINE_COLOR);
      g.drawPolygon(renderPolygon);
      final Rectangle bounds = renderPolygon.getBounds();
      final int x = bounds.x;
      final int y = bounds.y;
      final int w = bounds.width;
      final int h = bounds.height;
      final Graphics2D tmpGraphics = (Graphics2D) g.create();
      tmpGraphics.translate(x + w / 2, y + h / 2);
      tmpGraphics.rotate((double) polygon.getRotationAngle() / 180 * Math.PI);
      final String text = String.valueOf(i + 1);
      final int textWidth = tmpGraphics.getFontMetrics().stringWidth(text);
      final int ascent = tmpGraphics.getFontMetrics().getAscent();
      tmpGraphics.drawString(text, -textWidth / 2, ascent / 2);
      tmpGraphics.dispose();
    }
  }

  private Polygon transform(final Polygon polygon, final IRenderTransformation t) {
    final int[] xs = polygon.xpoints;
    final int[] ys = polygon.ypoints;
    final int[] newXs = t.transformX(xs);
    final int[] newYs = t.transformY(ys);
    return new Polygon(newXs, newYs, newXs.length);
  }

  private IRenderTransformation getRenderTransformation(final PcbMultiplySetup setup) {
    final int freeWidth = getSize().width - BORDER_SIZE - BORDER_SIZE;
    final int freeHeight = getSize().height - BORDER_SIZE - BORDER_SIZE;
    final double scaleX = (double) freeWidth / setup.getTargetBoardBounds().width;
    final double scaleY = (double) freeHeight / setup.getTargetBoardBounds().height;
    final double scaleFactor = Math.min(scaleX, scaleY);
    final IRenderTransformation t = new IRenderTransformation() {
      @Override
      public int transformX(final int worldX) {
        final int x = worldX - setup.getTargetBoardBounds().x;
        return BORDER_SIZE + (int) Math.round(x * scaleFactor);
      }

      @Override
      public int transformY(final int worldY) {
        final int y = worldY - setup.getTargetBoardBounds().y;
        return BORDER_SIZE + (int) Math.round(y * scaleFactor);
      }

      @Override
      public int[] transformX(final int[] worldXs) {
        final int[] values = new int[worldXs.length];
        for (int i = 0; i < values.length; ++i) {
          values[i] = transformX(worldXs[i]);
        }
        return values;
      }

      @Override
      public int[] transformY(final int[] worldYs) {
        final int[] values = new int[worldYs.length];
        for (int i = 0; i < values.length; ++i) {
          values[i] = transformY(worldYs[i]);
        }
        return values;
      }
    };
    return t;
  }
}