/**
 * Copyright (C) 2011 Markus Gebhard (mk@jave.de) and others
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package de.jave.kicad.pcbmultiplyer.swing;

import java.awt.Dimension;
import java.text.DecimalFormat;

import javax.swing.JComponent;
import javax.swing.JLabel;

import net.disy.commons.core.model.ObjectModel;
import net.disy.commons.core.model.listener.IChangeListener;
import net.disy.commons.core.util.Ensure;
import de.jave.kicad.pcbmultiplyer.MultiplyMode;
import de.jave.kicad.pcbmultiplyer.PcbMultiplySetup;
import de.jave.kicad.pcbmultiplyer.tree.ISolution;

public class StatusLine {

  private static final DecimalFormat PRICE_FORMAT = new DecimalFormat("#.#");
  private final JComponent content;
  private final JLabel label;
  private final ObjectModel<PcbMultiplySetup> setupModel;
  private final ObjectModel<ISolution> solutionModel;

  public StatusLine(
      final ObjectModel<PcbMultiplySetup> setupModel,
      final ObjectModel<ISolution> solutionModel) {
    Ensure.ensureArgumentNotNull(setupModel);
    Ensure.ensureArgumentNotNull(solutionModel);
    this.setupModel = setupModel;
    this.solutionModel = solutionModel;
    label = new JLabel();
    content = label;
    final IChangeListener updateTextListener = new IChangeListener() {
      @Override
      public void stateChanged() {
        updateText();
      }
    };
    setupModel.addChangeListener(updateTextListener);
    solutionModel.addChangeListener(updateTextListener);
    updateText();
  }

  private void updateText() {
    final PcbMultiplySetup setup = setupModel.getValue();
    final ISolution solution = solutionModel.getValue();
    if (setup == null || solution == null) {
      label.setText(""); //$NON-NLS-1$
      return;
    }
    final MultiplyMode mode = setup.getMode();
    final String boardCountText = getBoardCountText(solution);
    switch (mode) {
      case FIT_BOARD:
        final String efficiencyText = getEfficiencyText(setup, solution);
        final String pricePerBoardText = getPricePerBoardText(setup, solution);
        label.setText(boardCountText + ", " + efficiencyText + ". " + pricePerBoardText + ".");
        break;
      case FIXED_MULTIPLES:
        label.setText(boardCountText + ".");
        break;
      default:
        throw new UnsupportedOperationException();
    }
  }

  private String getPricePerBoardText(final PcbMultiplySetup setup, final ISolution solution) {
    final int pcbCount = solution.getPcbCount();
    final double price = (double) setup.getBoardPrice() / pcbCount;
    final String pricePerBoardText = "Price: "
        + PRICE_FORMAT.format(price)
        + " (ct/any unit) per board";
    return pricePerBoardText;
  }

  private String getBoardCountText(final ISolution solution) {
    final int pcbCount = solution.getPcbCount();
    final String boardCountText = pcbCount + " boards";
    return boardCountText;
  }

  private String getEfficiencyText(final PcbMultiplySetup setup, final ISolution solution) {
    final int pcbCount = solution.getPcbCount();
    final int spacing = setup.getSpacing();
    final Dimension size = setup.getOriginalContentBounds().getSize();
    final long pcbArea = size.width * size.height;
    final int minWidth = Math.min(size.width, size.height);
    final long optimumCovereredArea = pcbArea
        * pcbCount
        + Math.max(0, (pcbCount - 1))
        * minWidth
        * spacing;
    final long totalBoardArea = (long) setup.getTargetBoardBounds().width
        * (long) setup.getTargetBoardBounds().height;
    final double efficiencyPercent = 100.0 * optimumCovereredArea / totalBoardArea;
    return Math.round(efficiencyPercent) + "% efficiency";
  }

  public JComponent getContent() {
    return content;
  }
}