/**
 * Copyright (C) 2011 Markus Gebhard (mk@jave.de) and others
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package de.jave.kicad.pcbmultiplyer.swing.action;

import java.awt.Component;
import java.awt.Rectangle;
import java.lang.reflect.InvocationTargetException;

import net.disy.commons.core.message.Message;
import net.disy.commons.core.message.MessageType;
import net.disy.commons.core.model.ObjectModel;
import net.disy.commons.core.progress.ICancelable;
import net.disy.commons.core.progress.IInterruptableRunnableWithProgress;
import net.disy.commons.core.progress.IProgressMonitor;
import net.disy.commons.core.progress.ProgressUtilities;
import net.disy.commons.core.util.Ensure;
import net.disy.commons.swing.dialog.message.MessageDialogFactory;
import net.disy.commons.swing.dialog.progress.ProgressMonitorDialog;
import de.jave.kicad.pcb.model.KicadBoard;
import de.jave.kicad.pcbmultiplyer.MultiplyMode;
import de.jave.kicad.pcbmultiplyer.PcbMultiplyOptionsModel;
import de.jave.kicad.pcbmultiplyer.PcbMultiplySetup;
import de.jave.kicad.pcbmultiplyer.SolutionFactory;
import de.jave.kicad.pcbmultiplyer.SolutionRater;
import de.jave.kicad.pcbmultiplyer.tree.ISolution;
import de.jave.kicad.pcbmultiplyer.tree.MultiplySolution;
import de.jave.kicad.pcbmultiplyer.tree.PcbSolution;

public final class ApplyLayoutPerformer {

  private final ObjectModel<PcbMultiplySetup> setupModel;
  private final ObjectModel<ISolution> solutionModel;
  private final ObjectModel<KicadBoard> boardModel;
  private final PcbMultiplyOptionsModel optionsModel;

  public ApplyLayoutPerformer(
      final ObjectModel<PcbMultiplySetup> setupModel,
      final ObjectModel<ISolution> solutionModel,
      final ObjectModel<KicadBoard> boardModel,
      final PcbMultiplyOptionsModel optionsModel) {
    Ensure.ensureArgumentNotNull(setupModel);
    Ensure.ensureArgumentNotNull(solutionModel);
    Ensure.ensureArgumentNotNull(boardModel);
    Ensure.ensureArgumentNotNull(optionsModel);
    this.setupModel = setupModel;
    this.solutionModel = solutionModel;
    this.boardModel = boardModel;
    this.optionsModel = optionsModel;
  }

  public void execute(final Component parentComponent) {
    final KicadBoard board = boardModel.getValue();
    if (board == null) {
      return;
    }

    final Rectangle originalContentBounds = board.getPcbOutlineOrContentBounds();
    if (originalContentBounds == null) {
      MessageDialogFactory.showMessageDialog(parentComponent, new Message(
          "No outline segments found. Board must have pcb outline edges.",
          MessageType.WARNING));
      return;
    }

    final MultiplyMode mode = optionsModel.getMode();
    switch (mode) {
      case FIT_BOARD:
        performLayoutByFitBoard(parentComponent);
        break;
      case FIXED_MULTIPLES:
        performLayoutByFixedMultiples();
        break;
      default:
        throw new UnsupportedOperationException();
    }
  }

  private void performLayoutByFixedMultiples() {
    final KicadBoard board = boardModel.getValue();
    final Rectangle originalContentBounds = board.getPcbOutlineOrContentBounds();

    final int multiplesX = optionsModel.getMultiplesX();
    final int multiplesY = optionsModel.getMultiplesY();

    //TODO_NOW (gebhard) 23.01.2011: extract calculation methods! kicadToMm mmToKicad
    final int spacing = (int) Math.ceil(optionsModel.getSpacingMm() / 25.4 * 10000);

    final int targetBoardWidth = originalContentBounds.width
        * multiplesX
        + (multiplesX - 1)
        * spacing;
    final int targetBoardHeight = originalContentBounds.height
        * multiplesY
        + (multiplesY - 1)
        * spacing;
    final Rectangle targetBoardBounds = new Rectangle(
        originalContentBounds.x,
        originalContentBounds.y,
        targetBoardWidth,
        targetBoardHeight);

    final PcbMultiplySetup setup = new PcbMultiplySetup(
        MultiplyMode.FIXED_MULTIPLES,
        targetBoardBounds,
        originalContentBounds,
        spacing,
        optionsModel.getBoardPrice());
    setupModel.setValue(setup);

    final PcbSolution pcbSolution = new PcbSolution(originalContentBounds.getSize());
    final MultiplySolution solution = new MultiplySolution(
        pcbSolution,
        multiplesX,
        multiplesY,
        spacing);
    solutionModel.setValue(solution);
  }

  private void performLayoutByFitBoard(final Component parentComponent) {
    final KicadBoard board = boardModel.getValue();
    final Rectangle originalContentBounds = board.getPcbOutlineOrContentBounds();
    final int boardWidthMm = optionsModel.getBoardWidthMm();
    final int boardHeightMm = optionsModel.getBoardHeightMm();
    final double spacingMm = optionsModel.getSpacingMm();

    final int targetBoardWidth = (int) Math.floor(boardWidthMm / 25.4 * 10000);
    final int targetBoardHeight = (int) Math.floor(boardHeightMm / 25.4 * 10000);
    final Rectangle targetBoardBounds = new Rectangle(
        originalContentBounds.x,
        originalContentBounds.y,
        targetBoardWidth,
        targetBoardHeight);
    final int spacing = (int) Math.ceil(spacingMm / 25.4 * 10000);

    final PcbMultiplySetup setup = new PcbMultiplySetup(
        MultiplyMode.FIT_BOARD,
        targetBoardBounds,
        originalContentBounds,
        spacing,
        optionsModel.getBoardPrice());
    setupModel.setValue(setup);
    final SolutionRater solutionRater = new SolutionRater(setup);

    try {
      new ProgressMonitorDialog(parentComponent, "Finding optimal solution")
          .run(new IInterruptableRunnableWithProgress() {
            @Override
            public void run(final IProgressMonitor monitor, final ICancelable cancelable)
                throws InterruptedException,
                InvocationTargetException {
              final SolutionFactory solutionFactory = new SolutionFactory(setup, solutionRater);
              monitor.beginTaskWithUnknownTotalWork("Calculating solutions");
              ProgressUtilities.checkInterrupted(cancelable);
              final ISolution newSolution = solutionFactory.createBestSolutionWithDepth(cancelable);
              solutionModel.setValue(newSolution);
            }
          });
    }
    catch (final InterruptedException e) {
      return;
    }
    catch (final InvocationTargetException e) {
      MessageDialogFactory.showMessageDialog(parentComponent, new Message("Error occured.", e));
    }
  }
}