/**
 * Copyright (C) 2011 Markus Gebhard (mk@jave.de) and others
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package de.jave.kicad.pcbmultiplyer.swing.action;

import java.awt.Component;
import java.awt.Rectangle;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.List;

import net.disy.commons.core.io.FileUtilities;
import net.disy.commons.core.message.Message;
import net.disy.commons.core.model.ObjectModel;
import net.disy.commons.core.model.listener.IChangeListener;
import net.disy.commons.core.progress.ICancelable;
import net.disy.commons.core.progress.IInterruptableRunnableWithProgress;
import net.disy.commons.core.progress.IProgressMonitor;
import net.disy.commons.core.progress.ProgressUtilities;
import net.disy.commons.core.util.Ensure;
import net.disy.commons.swing.action.SmartAction;
import net.disy.commons.swing.dialog.message.MessageDialogFactory;
import net.disy.commons.swing.dialog.progress.ProgressMonitorDialog;
import net.disy.commons.swing.filechooser.SmartFileChooser;
import net.disy.commons.swing.filechooser.result.IFileChooserSaveResult;
import de.jave.kicad.pcb.model.KicadBoard;
import de.jave.kicad.pcb.model.KicadBoardContent;
import de.jave.kicad.pcb.model.PositioningTransformation;
import de.jave.kicad.pcb.persistence.write.KicadPcbFileWriter;
import de.jave.kicad.pcbmultiplyer.ContentTransformation;
import de.jave.kicad.pcbmultiplyer.swing.KicadBoardFileFileChooserConfiguration;
import de.jave.kicad.pcbmultiplyer.tree.ISolution;

public class SaveResultToNewBoardAction extends SmartAction {

  private final ObjectModel<KicadBoard> boardModel;
  private final ObjectModel<ISolution> solutionModel;

  public SaveResultToNewBoardAction(
      final ObjectModel<KicadBoard> boardModel,
      final ObjectModel<ISolution> solutionModel) {
    super("Save &result to new board...");
    Ensure.ensureArgumentNotNull(boardModel);
    Ensure.ensureArgumentNotNull(solutionModel);
    this.boardModel = boardModel;
    this.solutionModel = solutionModel;
    final IChangeListener updateEnabledListener = new IChangeListener() {
      @Override
      public void stateChanged() {
        updateEnabled();
      }
    };
    solutionModel.addChangeListener(updateEnabledListener);
    boardModel.addChangeListener(updateEnabledListener);
    updateEnabled();
  }

  private void updateEnabled() {
    setEnabled(boardModel.getValue() != null && solutionModel.getValue() != null);
  }

  @Override
  protected void execute(final Component parentComponent) {
    final KicadBoard originalBoard = boardModel.getValue();
    final ISolution solution = solutionModel.getValue();
    final File outputFile = createOutputFileNameSuggestion(originalBoard, solution);
    final KicadBoardFileFileChooserConfiguration configuration = new KicadBoardFileFileChooserConfiguration();
    final IFileChooserSaveResult result = SmartFileChooser.getInstance().performSaveFileChooser(
        parentComponent,
        configuration,
        outputFile,
        null);
    if (result.isCanceled()) {
      return;
    }
    final File file = result.getSelectedFile();
    final ProgressMonitorDialog dialog = new ProgressMonitorDialog(parentComponent, "Save");
    try {
      dialog.run(new IInterruptableRunnableWithProgress() {
        @Override
        public void run(final IProgressMonitor monitor, final ICancelable cancelable)
            throws InterruptedException,
            InvocationTargetException {
          monitor.beginTask("Saving board", 2 + solution.getPcbCount());
          monitor.subTask("Creating new board");
          final KicadBoard newBoard = originalBoard.clone();
          newBoard.setFile(file);
          monitor.worked(1);
          applyMultiplications(solution, newBoard, monitor, cancelable);
          ProgressUtilities.checkInterrupted(cancelable);
          monitor.subTask("Writing output file");
          try {
            KicadPcbFileWriter.write(newBoard);
          }
          catch (final IOException e) {
            throw new InvocationTargetException(e);
          }
          monitor.done();
        }
      });
    }
    catch (final InvocationTargetException e) {
      MessageDialogFactory.showMessageDialog(parentComponent, new Message(
          "Error writing output file.",
          e));
    }
    catch (final InterruptedException e) {
      return;
    }
  }

  private File createOutputFileNameSuggestion(
      final KicadBoard originalBoard,
      final ISolution solution) {
    final File originalFile = originalBoard.getFile();

    final File start = FileUtilities.getWithoutExtension(originalFile);
    final String extension = FileUtilities.getExtension(originalFile);

    final String suffix = "x" + solution.getPcbCount(); //$NON-NLS-1$
    final File outputFile = new File(start.getParentFile(), start.getName() + "_" //$NON-NLS-1$
        + suffix
        + "." //$NON-NLS-1$
        + extension);
    return outputFile;
  }

  private static void applyMultiplications(
      final ISolution solution,
      final KicadBoard board,
      final IProgressMonitor monitor,
      final ICancelable cancelable) throws InterruptedException {
    final KicadBoardContent content = board.getContent();
    final Rectangle originalContentBounds = board.getPcbOutlineOrContentBounds();
    board.removeContent();
    final List<ContentTransformation> transformations = solution.createTransformations();
    final List<PositioningTransformation> ts = ContentTransformation
        .convertToAffineTransformations(transformations, originalContentBounds);
    for (int i = 0; i < ts.size(); ++i) {
      ProgressUtilities.checkInterrupted(cancelable);
      monitor.subTask("Creating multiple " + (i + 1) + " of original content");
      final PositioningTransformation t = ts.get(i);
      final KicadBoardContent transformedContent = content.deriveTransformed(t);
      board.addContent(transformedContent);
      monitor.worked(1);
    }
  }
}